import json
import os
import re
import shutil
import sys

from pathlib import Path
from stat import S_ISFIFO

import rich_click as click
from dotmap import DotMap
from fp.fp import FreeProxy
from jinja2 import Template
from rich.live import Live
from rich.markdown import Markdown
from rich.rule import Rule
from rich.table import Table

from secator.config import CONFIG, ROOT_FOLDER, Config, default_config, config_path
from secator.decorators import OrderedGroup, register_runner
from secator.definitions import ADDONS_ENABLED, ASCII, DEV_PACKAGE, OPT_NOT_SUPPORTED, VERSION, STATE_COLORS
from secator.installer import ToolInstaller, fmt_health_table_row, get_health_table, get_version_info, get_distro_config
from secator.output_types import FINDING_TYPES, Info, Warning, Error
from secator.report import Report
from secator.rich import console
from secator.runners import Command, Runner
from secator.serializers.dataclass import loads_dataclass
from secator.template import TemplateLoader
from secator.utils import (
	debug, detect_host, discover_tasks, flatten, print_version, get_file_date,
	sort_files_by_date, get_file_timestamp, list_reports, get_info_from_report_path, human_to_timedelta
)

click.rich_click.USE_RICH_MARKUP = True

ALL_TASKS = discover_tasks()
ALL_CONFIGS = TemplateLoader.load_all()
ALL_WORKFLOWS = ALL_CONFIGS.workflow
ALL_SCANS = ALL_CONFIGS.scan
FINDING_TYPES_LOWER = [c.__name__.lower() for c in FINDING_TYPES]


#-----#
# CLI #
#-----#

@click.group(cls=OrderedGroup, invoke_without_command=True)
@click.option('--version', '-version', is_flag=True, default=False)
@click.pass_context
def cli(ctx, version):
	"""Secator CLI."""
	ctx.obj = {
		'piped_input': S_ISFIFO(os.fstat(0).st_mode),
		'piped_output': not sys.stdout.isatty()
	}
	if not ctx.obj['piped_output']:
		console.print(ASCII, highlight=False)
	if ctx.invoked_subcommand is None:
		if version:
			print_version()
		else:
			ctx.get_help()


#------#
# TASK #
#------#

@cli.group(aliases=['x', 't'])
@click.pass_context
def task(ctx):
	"""Run a task."""
	pass


for cls in ALL_TASKS:
	config = TemplateLoader(input={'name': cls.__name__, 'type': 'task'})
	register_runner(task, config)

#----------#
# WORKFLOW #
#----------#


@cli.group(cls=OrderedGroup, aliases=['w'])
@click.pass_context
def workflow(ctx):
	"""Run a workflow."""
	pass


for config in sorted(ALL_WORKFLOWS, key=lambda x: x['name']):
	register_runner(workflow, config)


#------#
# SCAN #
#------#

@cli.group(cls=OrderedGroup, aliases=['s'])
@click.pass_context
def scan(ctx):
	"""Run a scan."""
	pass


for config in sorted(ALL_SCANS, key=lambda x: x['name']):
	register_runner(scan, config)


#--------#
# WORKER #
#--------#

@cli.command(name='worker', context_settings=dict(ignore_unknown_options=True), aliases=['wk'])
@click.option('-n', '--hostname', type=str, default='runner', help='Celery worker hostname (unique).')
@click.option('-c', '--concurrency', type=int, default=100, help='Number of child processes processing the queue.')
@click.option('-r', '--reload', is_flag=True, help='Autoreload Celery on code changes.')
@click.option('-Q', '--queue', type=str, default='', help='Listen to a specific queue.')
@click.option('-P', '--pool', type=str, default='eventlet', help='Pool implementation.')
@click.option('--check', is_flag=True, help='Check if Celery worker is alive.')
@click.option('--dev', is_flag=True, help='Start a worker in dev mode (celery multi).')
@click.option('--stop', is_flag=True, help='Stop a worker in dev mode (celery multi).')
@click.option('--show', is_flag=True, help='Show command (celery multi).')
def worker(hostname, concurrency, reload, queue, pool, check, dev, stop, show):
	"""Run a worker."""

	# Check Celery addon is installed
	if not ADDONS_ENABLED['worker']:
		console.print(Error(message='Missing worker addon: please run "secator install addons worker".'))
		sys.exit(1)

	# Check broken / backend addon is installed
	broker_protocol = CONFIG.celery.broker_url.split('://')[0]
	backend_protocol = CONFIG.celery.result_backend.split('://')[0]
	if CONFIG.celery.broker_url:
		if (broker_protocol == 'redis' or backend_protocol == 'redis') and not ADDONS_ENABLED['redis']:
			console.print(Error(message='Missing redis addon: please run "secator install addons redis".'))
			sys.exit(1)

	# Debug Celery config
	from secator.celery import app, is_celery_worker_alive
	debug('conf', obj=dict(app.conf), obj_breaklines=True, sub='celery.app')
	debug('registered tasks', obj=list(app.tasks.keys()), obj_breaklines=True, sub='celery.app')

	if check:
		is_celery_worker_alive()
		return

	if not queue:
		queue = 'io,cpu,poll,' + ','.join(set([r['queue'] for r in app.conf.task_routes.values()]))

	app_str = 'secator.celery.app'
	celery = f'{sys.executable} -m celery'

	if dev:
		subcmd = 'stop' if stop else 'show' if show else 'start'
		logfile = '%n.log'
		pidfile = '%n.pid'
		queues = '-Q:1 celery -Q:2 io -Q:3 cpu'
		concur = '-c:1 10 -c:2 100 -c:3 4'
		pool = 'eventlet'
		cmd = f'{celery} -A {app_str} multi {subcmd} 3 {queues} -P {pool} {concur} --logfile={logfile} --pidfile={pidfile}'
	else:
		cmd = f'{celery} -A {app_str} worker -n {hostname} -Q {queue}'

	cmd += f' -P {pool}' if pool else ''
	cmd += f' -c {concurrency}' if concurrency else ''

	if reload:
		patterns = "celery.py;tasks/*.py;runners/*.py;serializers/*.py;output_types/*.py;hooks/*.py;exporters/*.py"
		cmd = f'watchmedo auto-restart --directory=./ --patterns="{patterns}" --recursive -- {cmd}'

	ret = Command.execute(cmd, name='secator_worker')
	sys.exit(ret.return_code)


#-------#
# UTILS #
#-------#


@cli.group(aliases=['u'])
def util():
	"""Run a utility."""
	pass


@util.command()
@click.option('--timeout', type=float, default=0.2, help='Proxy timeout (in seconds)')
@click.option('--number', '-n', type=int, default=1, help='Number of proxies')
def proxy(timeout, number):
	"""Get random proxies from FreeProxy."""
	if CONFIG.offline_mode:
		console.print(Error(message='Cannot run this command in offline mode.'))
		return
	proxy = FreeProxy(timeout=timeout, rand=True, anonym=True)
	for _ in range(number):
		url = proxy.get()
		console.print(url)


@util.command()
@click.argument('name', type=str, default=None, required=False)
@click.option('--host', '-h', type=str, default=None, help='Specify LHOST for revshell, otherwise autodetected.')
@click.option('--port', '-p', type=int, default=9001, show_default=True, help='Specify PORT for revshell')
@click.option('--interface', '-i', type=str, help='Interface to use to detect IP')
@click.option('--listen', '-l', is_flag=True, default=False, help='Spawn netcat listener on specified port')
@click.option('--force', is_flag=True)
def revshell(name, host, port, interface, listen, force):
	"""Show reverse shell source codes and run netcat listener (-l)."""
	if host is None:  # detect host automatically
		host = detect_host(interface)
		if not host:
			console.print(Error(message=f'Interface "{interface}" could not be found. Run "ifconfig" to see the list of available interfaces'))  # noqa: E501
			return
		else:
			console.print(Info(message=f'Detected host IP: {host}'))

	# Download reverse shells JSON from repo
	revshells_json = f'{CONFIG.dirs.revshells}/revshells.json'
	if not os.path.exists(revshells_json) or force:
		if CONFIG.offline_mode:
			console.print(Error(message='Cannot run this command in offline mode'))
			return
		ret = Command.execute(
			f'wget https://raw.githubusercontent.com/freelabz/secator/main/scripts/revshells.json && mv revshells.json {CONFIG.dirs.revshells}',  # noqa: E501
			cls_attributes={'shell': True}
		)
		if not ret.return_code == 0:
			sys.exit(1)

	# Parse JSON into shells
	with open(revshells_json) as f:
		shells = json.loads(f.read())
		for sh in shells:
			sh['alias'] = '_'.join(sh['name'].lower()
				.replace('-c', '')
				.replace('-e', '')
				.replace('-i', '')
				.replace('c#', 'cs')
				.replace('#', '')
				.replace('(', '')
				.replace(')', '')
				.strip()
				.split(' ')).replace('_1', '')
			cmd = re.sub(r"\s\s+", "", sh.get('command', ''), flags=re.UNICODE)
			cmd = cmd.replace('\n', ' ')
			sh['cmd_short'] = (cmd[:30] + '..') if len(cmd) > 30 else cmd

	shell = [
		shell for shell in shells if shell['name'] == name or shell['alias'] == name
	]
	if not shell:
		console.print('Available shells:', style='bold yellow')
		shells_str = [
			'[bold magenta]{alias:<20}[/][dim white]{name:<20}[/][dim gold3]{cmd_short:<20}[/]'.format(**sh)
			for sh in shells
		]
		console.print('\n'.join(shells_str))
	else:
		shell = shell[0]
		command = shell['command'].replace('[', r'\[')
		alias = shell['alias']
		name = shell['name']
		command_str = Template(command).render(ip=host, port=port, shell='bash')
		console.print(Rule(f'[bold gold3]{alias}[/] - [bold red]{name} REMOTE SHELL', style='bold red', align='left'))
		lang = shell.get('lang') or 'sh'
		if len(command.splitlines()) == 1:
			console.print(command_str, style='cyan', highlight=False, soft_wrap=True)
		else:
			md = Markdown(f'```{lang}\n{command_str}\n```')
			console.print(md)
			console.print(f'Save this script as rev.{lang} and run it on your target', style='dim italic')
		console.print()
		console.print(Rule(style='bold red'))

	if listen:
		console.print(Info(message=f'Starting netcat listener on port {port} ...'))
		cmd = f'nc -lvnp {port}'
		Command.execute(cmd)


@util.command()
@click.option('--directory', '-d', type=str, default=CONFIG.dirs.payloads, help='HTTP server directory')
@click.option('--host', '-h', type=str, default=None, help='HTTP host')
@click.option('--port', '-p', type=int, default=9001, help='HTTP server port')
@click.option('--interface', '-i', type=str, default=None, help='Interface to use to auto-detect host IP')
def serve(directory, host, port, interface):
	"""Run HTTP server to serve payloads."""
	console.print(Rule())
	console.print(f'Available payloads in {directory}: ', style='bold yellow')
	for fname in os.listdir(directory):
		if not host:
			host = detect_host(interface)
			if not host:
				console.print(Error(message=f'Interface "{interface}" could not be found. Run "ifconfig" to see the list of interfaces.'))  # noqa: E501
				return
		console.print(f'{fname} [dim][/]', style='bold magenta')
		console.print(f'wget http://{host}:{port}/{fname}', style='dim italic')
		console.print('')
	console.print(Rule())
	console.print(f'Started HTTP server on port {port}, waiting for incoming connections ...', style='bold yellow')
	Command.execute(f'{sys.executable} -m http.server {port}', cwd=directory)


@util.command()
@click.argument('record_name', type=str, default=None)
@click.option('--script', '-s', type=str, default=None, help='Script to run. See scripts/stories/ for examples.')
@click.option('--interactive', '-i', is_flag=True, default=False, help='Interactive record.')
@click.option('--width', '-w', type=int, default=None, help='Recording width')
@click.option('--height', '-h', type=int, default=None, help='Recording height')
@click.option('--output-dir', type=str, default=f'{ROOT_FOLDER}/images')
def record(record_name, script, interactive, width, height, output_dir):
	"""Record secator session using asciinema."""
	# 120 x 30 is a good ratio for GitHub
	width = width or console.size.width
	height = height or console.size.height
	attrs = {
		'shell': False,
		'env': {
			'RECORD': '1',
			'LINES': str(height),
			'PS1': '$ ',
			'COLUMNS': str(width),
			'TERM': 'xterm-256color'
		}
	}
	output_cast_path = f'{output_dir}/{record_name}.cast'
	output_gif_path = f'{output_dir}/{record_name}.gif'

	# Run automated 'story' script with asciinema-automation
	if script:
		# If existing cast file, remove it
		if os.path.exists(output_cast_path):
			os.unlink(output_cast_path)
			console.print(Info(message=f'Removed existing {output_cast_path}'))

		with console.status(Info(message='Recording with asciinema ...')):
			Command.execute(
				f'asciinema-automation -aa "-c /bin/sh" {script} {output_cast_path} --timeout 200',
				cls_attributes=attrs,
				raw=True,
			)
			console.print(f'Generated {output_cast_path}', style='bold green')
	elif interactive:
		os.environ.update(attrs['env'])
		Command.execute(f'asciinema rec -c /bin/bash --stdin --overwrite {output_cast_path}')

	# Resize cast file
	if os.path.exists(output_cast_path):
		with console.status('[bold gold3]Cleaning up .cast and set custom settings ...'):
			with open(output_cast_path, 'r') as f:
				lines = f.readlines()
			updated_lines = []
			for ix, line in enumerate(lines):
				tmp_line = json.loads(line)
				if ix == 0:
					tmp_line['width'] = width
					tmp_line['height'] = height
					tmp_line['env']['SHELL'] = '/bin/sh'
					lines[0] = json.dumps(tmp_line) + '\n'
					updated_lines.append(json.dumps(tmp_line) + '\n')
				elif tmp_line[2].endswith(' \r'):
					tmp_line[2] = tmp_line[2].replace(' \r', '')
					updated_lines.append(json.dumps(tmp_line) + '\n')
				else:
					updated_lines.append(line)
			with open(output_cast_path, 'w') as f:
				f.writelines(updated_lines)
			console.print('')

		# Edit cast file to reduce long timeouts
		with console.status('[bold gold3] Editing cast file to reduce long commands ...'):
			Command.execute(
				f'asciinema-edit quantize --range 1 {output_cast_path} --out {output_cast_path}.tmp',
				cls_attributes=attrs,
				raw=True,
			)
			if os.path.exists(f'{output_cast_path}.tmp'):
				os.replace(f'{output_cast_path}.tmp', output_cast_path)
			console.print(f'Edited {output_cast_path}', style='bold green')

	# Convert to GIF
	with console.status(f'[bold gold3]Converting to {output_gif_path} ...[/]'):
		Command.execute(
			f'agg {output_cast_path} {output_gif_path}',
			cls_attributes=attrs,
		)
		console.print(Info(message=f'Generated {output_gif_path}'))


@util.command('build')
@click.option('--version', type=str, help='Override version specified in pyproject.toml')
def build(version):
	"""Build secator PyPI package."""
	if not DEV_PACKAGE:
		console.print(Error(message='You MUST use a development version of secator to make builds'))
		sys.exit(1)
	if not ADDONS_ENABLED['build']:
		console.print(Error(message='Missing build addon: please run "secator install addons build"'))
		sys.exit(1)

	# Update version in pyproject.toml if --version is explicitely passed
	if version:
		pyproject_toml_path = Path.cwd() / 'pyproject.toml'
		if not pyproject_toml_path.exists():
			console.print(Error(message='You must be in the secator root directory to make builds with --version'))
			sys.exit(1)
		console.print(Info(message=f'Updating version in pyproject.toml to {version}'))
		with open(pyproject_toml_path, "r") as file:
			content = file.read()
		updated_content = re.sub(r'^\s*version\s*=\s*".*?"', f'version = "{version}"', content, flags=re.MULTILINE)
		with open(pyproject_toml_path, "w") as file:
			file.write(updated_content)

	with console.status('[bold gold3]Building PyPI package...[/]'):
		ret = Command.execute(f'{sys.executable} -m hatch build', name='hatch build', cwd=ROOT_FOLDER)
		sys.exit(ret.return_code)


@util.command('publish')
def publish():
	"""Publish secator PyPI package."""
	if not DEV_PACKAGE:
		console.print(Error(message='You MUST use a development version of secator to publish builds.'))
		sys.exit(1)
	if not ADDONS_ENABLED['build']:
		console.print(Error(message='Missing build addon: please run "secator install addons build"'))
		sys.exit(1)
	os.environ['HATCH_INDEX_USER'] = '__token__'
	hatch_token = os.environ.get('HATCH_INDEX_AUTH')
	if not hatch_token:
		console.print(Error(message='Missing PyPI auth token (HATCH_INDEX_AUTH env variable).'))
		sys.exit(1)
	with console.status('[bold gold3]Publishing PyPI package...[/]'):
		ret = Command.execute(f'{sys.executable} -m hatch publish', name='hatch publish', cwd=ROOT_FOLDER)
		sys.exit(ret.return_code)


#--------#
# CONFIG #
#--------#

@cli.group(aliases=['c'])
def config():
	"""View or edit config."""
	pass


@config.command('get')
@click.option('--full', is_flag=True, help='Show full config (with defaults)')
@click.argument('key', required=False)
def config_get(full, key=None):
	"""Get config value."""
	if key is None:
		partial = not full and CONFIG != default_config
		CONFIG.print(partial=partial)
		return
	CONFIG.get(key)


@config.command('set')
@click.argument('key')
@click.argument('value')
def config_set(key, value):
	"""Set config value."""
	CONFIG.set(key, value)
	config = CONFIG.validate()
	if config:
		CONFIG.get(key)
		saved = CONFIG.save()
		if not saved:
			return
		console.print(f'[bold green]:tada: Saved config to [/]{CONFIG._path}')
	else:
		console.print(Error(message='Invalid config, not saving it.'))


@config.command('edit')
@click.option('--resume', is_flag=True)
def config_edit(resume):
	"""Edit config."""
	tmp_config = CONFIG.dirs.data / 'config.yml.patch'
	if not tmp_config.exists() or not resume:
		shutil.copyfile(config_path, tmp_config)
	click.edit(filename=tmp_config)
	config = Config.parse(path=tmp_config)
	if config:
		config.save(config_path)
		console.print(f'\n[bold green]:tada: Saved config to [/]{config_path}.')
		tmp_config.unlink()
	else:
		console.print('\n[bold green]Hint:[/] Run "secator config edit --resume" to edit your patch and fix issues.')


@config.command('default')
@click.option('--save', type=str, help='Save default config to file.')
def config_default(save):
	"""Get default config."""
	default_config.print(partial=False)
	if save:
		default_config.save(target_path=Path(save), partial=False)
		console.print(f'\n[bold green]:tada: Saved default config to [/]{save}.')


# TODO: implement reset method
# @_config.command('reset')
# @click.argument('key')
# def config_reset(key):
# 	"""Reset config value to default."""
# 	success = CONFIG.set(key, None)
# 	if success:
# 		CONFIG.print()
# 		CONFIG.save()
# 		console.print(f'\n[bold green]:tada: Saved config to [/]{CONFIG._path}')

#-----------#
# WORKSPACE #
#-----------#
@cli.group(aliases=['ws'])
def workspace():
	"""Workspaces."""
	pass


@workspace.command('list')
def workspace_list():
	"""List workspaces."""
	workspaces = {}
	json_reports = []
	for root, _, files in os.walk(CONFIG.dirs.reports):
		for file in files:
			if file.endswith('report.json'):
				path = Path(root) / file
				json_reports.append(path)
	json_reports = sorted(json_reports, key=lambda x: x.stat().st_mtime, reverse=False)
	for path in json_reports:
		ws, runner_type, number = str(path).split('/')[-4:-1]
		if ws not in workspaces:
			workspaces[ws] = {'count': 0, 'path': '/'.join(str(path).split('/')[:-3])}
		workspaces[ws]['count'] += 1

	# Build table
	table = Table()
	table.add_column("Workspace name", style="bold gold3")
	table.add_column("Run count", overflow='fold')
	table.add_column("Path")
	for workspace, config in workspaces.items():
		table.add_row(workspace, str(config['count']), config['path'])
	console.print(table)


#--------#
# REPORT #
#--------#


@cli.group(aliases=['r'])
def report():
	"""Reports."""
	pass


@report.command('show')
@click.argument('report_query', required=False)
@click.option('-o', '--output', type=str, default='console', help='Exporters')
@click.option('-r', '--runner-type', type=str, default=None, help='Filter by runner type. Choices: task, workflow, scan')  # noqa: E501
@click.option('-d', '--time-delta', type=str, default=None, help='Keep results newer than time delta. E.g: 26m, 1d, 1y')  # noqa: E501
@click.option('-t', '--type', type=str, default='', help=f'Filter by output type. Choices: {FINDING_TYPES_LOWER}')
@click.option('-q', '--query', type=str, default=None, help='Query results using a Python expression')
@click.option('-w', '-ws', '--workspace', type=str, default=None, help='Filter by workspace name')
@click.option('-u', '--unified', is_flag=True, default=False, help='Show unified results (merge reports and de-duplicates results)')  # noqa: E501
def report_show(report_query, output, runner_type, time_delta, type, query, workspace, unified):
	"""Show report results and filter on them."""

	# Get extractors
	otypes = [o.__name__.lower() for o in FINDING_TYPES]
	extractors = []
	if type:
		type = type.split(',')
		for typedef in type:
			if typedef:
				if '.' in typedef:
					_type, _field = tuple(typedef.split('.'))
				else:
					_type = typedef
					_field = None
				extractors.append({
					'type': _type,
					'field': _field,
					'condition': query or 'True'
				})
	elif query:
		query = query.split(';')
		for part in query:
			_type = part.split('.')[0]
			if _type in otypes:
				part = part.replace(_type, 'item')
				extractor = {
					'type': _type,
					'condition': part or 'True'
				}
				extractors.append(extractor)

	# Build runner instance
	current = get_file_timestamp()
	runner = DotMap({
		"config": {
			"name": f"consolidated_report_{current}"
		},
		"name": "runner",
		"workspace_name": "_consolidated",
		"reports_folder": Path.cwd(),
	})
	exporters = Runner.resolve_exporters(output)

	# Build report queries from fuzzy input
	paths = []
	if report_query:
		report_query = report_query.split(',')
	else:
		report_query = []

	# Load all report paths
	load_all_reports = any([not Path(p).exists() for p in report_query])
	all_reports = []
	if load_all_reports or workspace:
		all_reports = list_reports(workspace=workspace, type=runner_type, timedelta=human_to_timedelta(time_delta))
	if not report_query:
		report_query = all_reports

	for query in report_query:
		query = str(query)
		if not query.endswith('/'):
			query += '/'
		path = Path(query)
		if not path.exists():
			matches = []
			for path in all_reports:
				if query in str(path):
					matches.append(path)
			if not matches:
				console.print(
					f'[bold orange3]Query {query} did not return any matches. [/][bold green]Ignoring.[/]')
			paths.extend(matches)
		else:
			paths.append(path)
	paths = sort_files_by_date(paths)

	# Load reports, extract results
	all_results = []
	for ix, path in enumerate(paths):
		if unified:
			console.print(rf'Loading {path} \[[bold yellow4]{ix + 1}[/]/[bold yellow4]{len(paths)}[/]] \[results={len(all_results)}]...')  # noqa: E501
		with open(path, 'r') as f:
			data = loads_dataclass(f.read())
			try:
				info = get_info_from_report_path(path)
				runner_type = info.get('type', 'unknowns')[:-1]
				runner.results = flatten(list(data['results'].values()))
				if unified:
					all_results.extend(runner.results)
					continue
				report = Report(runner, title=f"Consolidated report - {current}", exporters=exporters)
				report.build(extractors=extractors if not unified else [])
				file_date = get_file_date(path)
				runner_name = data['info']['name']
				console.print(
					f'\n{path} ([bold blue]{runner_name}[/] [dim]{runner_type}[/]) ([dim]{file_date}[/]):')
				if report.is_empty():
					if len(paths) == 1:
						console.print(Warning(message='No results in report.'))
					else:
						console.print(Warning(message='No new results since previous scan.'))
					continue
				report.send()
			except json.decoder.JSONDecodeError as e:
				console.print(Error(message=f'Could not load {path}: {str(e)}'))

	if unified:
		console.print(f'\n:wrench: [bold gold3]Building report by crunching {len(all_results)} results ...[/]')
		console.print(':coffee: [dim]Note that this can take a while when the result count is high...[/]')
		runner.results = all_results
		report = Report(runner, title=f"Consolidated report - {current}", exporters=exporters)
		report.build(extractors=extractors, dedupe=True)
		report.send()


@report.command('list')
@click.option('-ws', '-w', '--workspace', type=str)
@click.option('-r', '--runner-type', type=str, default=None, help='Filter by runner type. Choices: task, workflow, scan')  # noqa: E501
@click.option('-d', '--time-delta', type=str, default=None, help='Keep results newer than time delta. E.g: 26m, 1d, 1y')  # noqa: E501
def report_list(workspace, runner_type, time_delta):
	"""List all secator reports."""
	paths = list_reports(workspace=workspace, type=runner_type, timedelta=human_to_timedelta(time_delta))
	paths = sorted(paths, key=lambda x: x.stat().st_mtime, reverse=False)

	# Build table
	table = Table()
	table.add_column("Workspace", style="bold gold3")
	table.add_column("Path", overflow='fold')
	table.add_column("Name")
	table.add_column("Id")
	table.add_column("Date")
	table.add_column("Status", style="green")

	# Load each report
	for path in paths:
		try:
			info = get_info_from_report_path(path)
			with open(path, 'r') as f:
				content = json.loads(f.read())
			data = {
				'workspace': info['workspace'],
				'name': f"[bold blue]{content['info']['name']}[/]",
				'status': content['info'].get('status', ''),
				'id': info['type'] + '/' + info['id'],
				'date': get_file_date(path),  # Assuming get_file_date returns a readable date
			}
			status_color = STATE_COLORS[data['status']] if data['status'] in STATE_COLORS else 'white'

			# Update table
			table.add_row(
				data['workspace'],
				str(path),
				data['name'],
				data['id'],
				data['date'],
				f"[{status_color}]{data['status']}[/]"
			)
		except json.JSONDecodeError as e:
			console.print(Error(message=f'Could not load {path}: {str(e)}'))

	if len(paths) > 0:
		console.print(table)
	else:
		console.print(Error(message='No results found.'))


@report.command('export')
@click.argument('json_path', type=str)
@click.option('--output-folder', '-of', type=str)
@click.option('-output', '-o', type=str)
def report_export(json_path, output_folder, output):
	with open(json_path, 'r') as f:
		data = loads_dataclass(f.read())

	runner_instance = DotMap({
		"config": {
			"name": data['info']['name']
		},
		"workspace_name": json_path.split('/')[-4],
		"reports_folder": output_folder or Path.cwd(),
		"data": data,
		"results": flatten(list(data['results'].values()))
	})
	exporters = Runner.resolve_exporters(output)
	report = Report(runner_instance, title=data['info']['title'], exporters=exporters)
	report.data = data
	report.send()


#--------#
# DEPLOY #
#--------#

# TODO: work on this
# @cli.group(aliases=['d'])
# def deploy():
# 	"""Deploy secator."""
# 	pass

# @deploy.command()
# def docker_compose():
# 	"""Deploy secator on docker-compose."""
# 	pass

# @deploy.command()
# @click.option('-t', '--target', type=str, default='minikube', help='Deployment target amongst minikube, gke')
# def k8s():
# 	"""Deploy secator on Kubernetes."""
# 	pass


#--------#
# HEALTH #
#--------#

@cli.command(name='health')
@click.option('--json', '-json', is_flag=True, default=False, help='JSON lines output')
@click.option('--debug', '-debug', is_flag=True, default=False, help='Debug health output')
@click.option('--strict', '-strict', is_flag=True, default=False, help='Fail if missing tools')
def health(json, debug, strict):
	"""[dim]Get health status.[/]"""
	tools = ALL_TASKS
	status = {'secator': {}, 'languages': {}, 'tools': {}, 'addons': {}}

	# Check secator
	console.print(':wrench: [bold gold3]Checking secator ...[/]')
	info = get_version_info('secator', '-version', 'freelabz/secator')
	table = get_health_table()
	with Live(table, console=console):
		row = fmt_health_table_row(info)
		table.add_row(*row)
	status['secator'] = info

	# Check addons
	console.print('\n:wrench: [bold gold3]Checking installed addons ...[/]')
	table = get_health_table()
	with Live(table, console=console):
		for addon, installed in ADDONS_ENABLED.items():
			info = {
				'name': addon,
				'version': None,
				'status': 'ok' if installed else 'missing',
				'latest_version': None,
				'installed': installed,
				'location': None
			}
			row = fmt_health_table_row(info, 'addons')
			table.add_row(*row)
			status['addons'][addon] = info

	# Check languages
	console.print('\n:wrench: [bold gold3]Checking installed languages ...[/]')
	version_cmds = {'go': 'version', 'python3': '--version', 'ruby': '--version'}
	table = get_health_table()
	with Live(table, console=console):
		for lang, version_flag in version_cmds.items():
			info = get_version_info(lang, version_flag)
			row = fmt_health_table_row(info, 'langs')
			table.add_row(*row)
			status['languages'][lang] = info

	# Check tools
	console.print('\n:wrench: [bold gold3]Checking installed tools ...[/]')
	table = get_health_table()
	with Live(table, console=console):
		for tool in tools:
			info = get_version_info(
				tool.cmd.split(' ')[0],
				tool.version_flag or f'{tool.opt_prefix}version',
				tool.install_github_handle,
				tool.install_cmd
			)
			row = fmt_health_table_row(info, 'tools')
			table.add_row(*row)
			status['tools'][tool.__name__] = info
	console.print('')

	# Print JSON health
	if json:
		import json as _json
		print(_json.dumps(status))

	# Strict mode
	if strict:
		error = False
		for tool, info in status['tools'].items():
			if not info['installed']:
				console.print(Error(message=f'{tool} not installed and strict mode is enabled.'))
				error = True
		if error:
			sys.exit(1)
		console.print(Info(message='Strict healthcheck passed !'))


#---------#
# INSTALL #
#---------#


def run_install(title=None, cmd=None, packages=None, next_steps=None):
	if CONFIG.offline_mode:
		console.print(Error(message='Cannot run this command in offline mode.'))
		return
	with console.status(f'[bold yellow] Installing {title}...'):
		if cmd:
			from secator.installer import SourceInstaller
			status = SourceInstaller.install(cmd)
		elif packages:
			from secator.installer import PackageInstaller
			status = PackageInstaller.install(packages)
		return_code = 1
		if status.is_ok():
			return_code = 0
			if next_steps:
				console.print('[bold gold3]:wrench: Next steps:[/]')
				for ix, step in enumerate(next_steps):
					console.print(f'   :keycap_{ix}: {step}')
		sys.exit(return_code)


@cli.group()
def install():
	"""[dim]Install langs, tools and addons.[/]"""
	pass


@install.group()
def addons():
	"Install addons."
	pass


@addons.command('worker')
def install_worker():
	"Install Celery worker addon."
	run_install(
		cmd=f'{sys.executable} -m pip install secator[worker]',
		title='Celery worker addon',
		next_steps=[
			'Run [bold green4]secator worker[/] to run a Celery worker using the file system as a backend and broker.',
			'Run [bold green4]secator x httpx testphp.vulnweb.com[/] to admire your task running in a worker.',
			r'[dim]\[optional][/dim] Run [bold green4]secator install addons redis[/] to setup Redis backend / broker.'
		]
	)


@addons.command('gdrive')
def install_gdrive():
	"Install Google Drive addon."
	run_install(
		cmd=f'{sys.executable} -m pip install secator[google]',
		title='Google Drive addon',
		next_steps=[
			'Run [bold green4]secator config set addons.gdrive.credentials_path <VALUE>[/].',
			'Run [bold green4]secator config set addons.gdrive.drive_parent_folder_id <VALUE>[/].',
			'Run [bold green4]secator x httpx testphp.vulnweb.com -o gdrive[/] to send reports to Google Drive.'
		]
	)


@addons.command('gcs')
def install_gcs():
	"Install Google Cloud Storage addon."
	run_install(
		cmd=f'{sys.executable} -m pip install secator[gcs]',
		title='Google Cloud Storage addon',
		next_steps=[
			'Run [bold green4]secator config set addons.gcs.credentials_path <VALUE>[/].',
		]
	)


@addons.command('mongodb')
def install_mongodb():
	"Install MongoDB addon."
	run_install(
		cmd=f'{sys.executable} -m pip install secator[mongodb]',
		title='MongoDB addon',
		next_steps=[
			r'[dim]\[optional][/] Run [bold green4]docker run --name mongo -p 27017:27017 -d mongo:latest[/] to run a local MongoDB instance.',  # noqa: E501
			'Run [bold green4]secator config set addons.mongodb.url mongodb://<URL>[/].',
			'Run [bold green4]secator x httpx testphp.vulnweb.com -driver mongodb[/] to save results to MongoDB.'
		]
	)


@addons.command('redis')
def install_redis():
	"Install Redis addon."
	run_install(
		cmd=f'{sys.executable} -m pip install secator[redis]',
		title='Redis addon',
		next_steps=[
			r'[dim]\[optional][/] Run [bold green4]docker run --name redis -p 6379:6379 -d redis[/] to run a local Redis instance.',  # noqa: E501
			'Run [bold green4]secator config set celery.broker_url redis://<URL>[/]',
			'Run [bold green4]secator config set celery.result_backend redis://<URL>[/]',
			'Run [bold green4]secator worker[/] to run a worker.',
			'Run [bold green4]secator x httpx testphp.vulnweb.com[/] to run a test task.'
		]
	)


@addons.command('dev')
def install_dev():
	"Install dev addon."
	run_install(
		cmd=f'{sys.executable} -m pip install secator[dev]',
		title='dev addon',
		next_steps=[
			'Run [bold green4]secator test lint[/] to run lint tests.',
			'Run [bold green4]secator test unit[/] to run unit tests.',
			'Run [bold green4]secator test integration[/] to run integration tests.',
		]
	)


@addons.command('trace')
def install_trace():
	"Install trace addon."
	run_install(
		cmd=f'{sys.executable} -m pip install secator[trace]',
		title='trace addon',
		next_steps=[
		]
	)


@addons.command('build')
def install_build():
	"Install build addon."
	run_install(
		cmd=f'{sys.executable} -m pip install secator[build]',
		title='build addon',
		next_steps=[
			'Run [bold green4]secator u build pypi[/] to build the PyPI package.',
			'Run [bold green4]secator u publish pypi[/] to publish the PyPI package.',
			'Run [bold green4]secator u build docker[/] to build the Docker image.',
			'Run [bold green4]secator u publish docker[/] to publish the Docker image.',
		]
	)


@install.group()
def langs():
	"Install languages."
	pass


@langs.command('go')
def install_go():
	"""Install Go."""
	run_install(
		cmd='wget -O - https://raw.githubusercontent.com/freelabz/secator/main/scripts/install_go.sh | sudo sh',
		title='Go',
		next_steps=[
			'Add ~/go/bin to your $PATH'
		]
	)


@langs.command('ruby')
def install_ruby():
	"""Install Ruby."""
	run_install(
		packages={
			'apt': ['ruby-full', 'rubygems'],
			'apk': ['ruby', 'ruby-dev'],
			'pacman': ['ruby', 'ruby-dev'],
			'brew': ['ruby']
		},
		title='Ruby'
	)


@install.command('tools')
@click.argument('cmds', required=False)
@click.option('--cleanup', is_flag=True, default=False)
def install_tools(cmds, cleanup):
	"""Install supported tools."""
	if CONFIG.offline_mode:
		console.print(Error(message='Cannot run this command in offline mode.'))
		return
	if cmds is not None:
		cmds = cmds.split(',')
		tools = [cls for cls in ALL_TASKS if cls.__name__ in cmds]
	else:
		tools = ALL_TASKS
	tools.sort(key=lambda x: x.__name__)
	return_code = 0
	if not tools:
		cmd_str = ' '.join(cmds)
		console.print(Error(message=f'No tools found for {cmd_str}.'))
		return
	for ix, cls in enumerate(tools):
		with console.status(f'[bold yellow][{ix + 1}/{len(tools)}] Installing {cls.__name__} ...'):
			status = ToolInstaller.install(cls)
			if not status.is_ok():
				return_code = 1
		console.print()
	if cleanup:
		distro = get_distro_config()
		cleanup_cmds = [
			'go clean -cache',
			'go clean -modcache',
			'pip cache purge',
			'gem cleanup --user-install',
			'gem clean --user-install',
		]
		if distro.pm_finalizer:
			cleanup_cmds.append(f'sudo {distro.pm_finalizer}')
		cmd = ' && '.join(cleanup_cmds)
		Command.execute(cmd, cls_attributes={'shell': True}, quiet=False)
	sys.exit(return_code)


#--------#
# UPDATE #
#--------#

@cli.command('update')
@click.option('--all', '-a', is_flag=True, help='Update all secator dependencies (addons, tools, ...)')
def update(all):
	"""[dim]Update to latest version.[/]"""
	if CONFIG.offline_mode:
		console.print(Error(message='Cannot run this command in offline mode.'))
		sys.exit(1)

	# Check current and latest version
	info = get_version_info('secator', '-version', 'freelabz/secator', version=VERSION)
	latest_version = info['latest_version']
	do_update = True

	# Skip update if latest
	if info['status'] == 'latest':
		console.print(Info(message=f'secator is already at the newest version {latest_version} !'))
		do_update = False

	# Fail if unknown latest
	if not latest_version:
		console.print(Error(message='Could not fetch latest secator version.'))
		sys.exit(1)

	# Update secator
	if do_update:
		console.print(f'[bold gold3]:wrench: Updating secator from {VERSION} to {latest_version} ...[/]')
		if 'pipx' in sys.executable:
			ret = Command.execute(f'pipx install secator=={latest_version} --force')
		else:
			ret = Command.execute(f'pip install secator=={latest_version}')
		if not ret.return_code == 0:
			sys.exit(1)

	# Update tools
	if all:
		return_code = 0
		for cls in ALL_TASKS:
			cmd = cls.cmd.split(' ')[0]
			version_flag = cls.version_flag or f'{cls.opt_prefix}version'
			version_flag = None if cls.version_flag == OPT_NOT_SUPPORTED else version_flag
			info = get_version_info(cmd, version_flag, cls.install_github_handle)
			if not info['installed'] or info['status'] == 'outdated' or not info['latest_version']:
				with console.status(f'[bold yellow]Installing {cls.__name__} ...'):
					status = ToolInstaller.install(cls)
					if not status.is_ok():
						return_code = 1
		sys.exit(return_code)

#-------#
# ALIAS #
#-------#


@cli.group()
def alias():
	"""[dim]Configure aliases.[/]"""
	pass


@alias.command('enable')
@click.pass_context
def enable_aliases(ctx):
	"""Enable aliases."""
	fpath = f'{CONFIG.dirs.data}/.aliases'
	aliases = ctx.invoke(list_aliases, silent=True)
	aliases_str = '\n'.join(aliases)
	with open(fpath, 'w') as f:
		f.write(aliases_str)
	console.print('')
	console.print(f':file_cabinet: Alias file written to {fpath}', style='bold green')
	console.print('To load the aliases, run:')
	md = f"""
```sh
source {fpath}                     # load the aliases in the current shell
echo "source {fpath} >> ~/.bashrc" # or add this line to your ~/.bashrc to load them automatically
```
"""
	console.print(Markdown(md))
	console.print()


@alias.command('disable')
@click.pass_context
def disable_aliases(ctx):
	"""Disable aliases."""
	fpath = f'{CONFIG.dirs.data}/.unalias'
	aliases = ctx.invoke(list_aliases, silent=True)
	aliases_str = ''
	for alias in aliases:
		aliases_str += alias.split('=')[0].replace('alias', 'unalias') + '\n'
	console.print(f':file_cabinet: Unalias file written to {fpath}', style='bold green')
	console.print('To unload the aliases, run:')
	with open(fpath, 'w') as f:
		f.write(aliases_str)
	md = f"""
```sh
source {fpath}
```
"""
	console.print(Markdown(md))
	console.print()


@alias.command('list')
@click.option('--silent', is_flag=True, default=False, help='No print')
def list_aliases(silent):
	"""List aliases"""
	aliases = []
	aliases.extend([
		f'alias {task.__name__}="secator x {task.__name__}"'
		for task in ALL_TASKS
	])
	aliases.extend([
		f'alias {workflow.alias}="secator w {workflow.name}"'
		for workflow in ALL_WORKFLOWS
	])
	aliases.extend([
		f'alias {workflow.name}="secator w {workflow.name}"'
		for workflow in ALL_WORKFLOWS
	])
	aliases.extend([
		f'alias scan_{scan.name}="secator s {scan.name}"'
		for scan in ALL_SCANS
	])
	aliases.append('alias listx="secator x"')
	aliases.append('alias listw="secator w"')
	aliases.append('alias lists="secator s"')

	if silent:
		return aliases
	console.print('Aliases:')
	for alias in aliases:
		alias_split = alias.split('=')
		alias_name, alias_cmd = alias_split[0].replace('alias ', ''), alias_split[1].replace('"', '')
		console.print(f'[bold magenta]{alias_name:<15}-> {alias_cmd}')

	return aliases


#------#
# TEST #
#------#


@cli.group(cls=OrderedGroup)
def test():
	"""[dim]Run tests."""
	if not DEV_PACKAGE:
		console.print(Error(message='You MUST use a development version of secator to run tests.'))
		sys.exit(1)
	if not ADDONS_ENABLED['dev']:
		console.print(Error(message='Missing dev addon: please run "secator install addons dev"'))
		sys.exit(1)
	pass


def run_test(cmd, name):
	"""Run a test and return the result.

	Args:
		cmd: Command to run.
		name: Name of the test.
	"""
	result = Command.execute(cmd, name=name + ' tests', cwd=ROOT_FOLDER)
	if result.return_code == 0:
		console.print(f':tada: {name.capitalize()} tests passed !', style='bold green')
	sys.exit(result.return_code)


@test.command()
def lint():
	"""Run lint tests."""
	cmd = f'{sys.executable} -m flake8 secator/'
	run_test(cmd, 'lint')


@test.command()
@click.option('--tasks', type=str, default='', help='Secator tasks to test (comma-separated)')
@click.option('--workflows', type=str, default='', help='Secator workflows to test (comma-separated)')
@click.option('--scans', type=str, default='', help='Secator scans to test (comma-separated)')
@click.option('--test', '-t', type=str, help='Secator test to run')
def unit(tasks, workflows, scans, test):
	"""Run unit tests."""
	os.environ['TEST_TASKS'] = tasks or ''
	os.environ['TEST_WORKFLOWS'] = workflows or ''
	os.environ['TEST_SCANS'] = scans or ''
	os.environ['SECATOR_DIRS_DATA'] = '/tmp/.secator'
	os.environ['SECATOR_OFFLINE_MODE'] = "1"
	os.environ['SECATOR_HTTP_STORE_RESPONSES'] = '0'
	os.environ['SECATOR_RUNNERS_SKIP_CVE_SEARCH'] = '1'

	import shutil
	shutil.rmtree('/tmp/.secator', ignore_errors=True)
	cmd = f'{sys.executable} -m coverage run --omit="*test*" --data-file=.coverage.unit -m pytest -s -v tests/unit'
	if test:
		test_str = ' or '.join(test.split(','))
		cmd += f' -k "{test_str}"'
	run_test(cmd, 'unit')


@test.command()
@click.option('--tasks', type=str, default='', help='Secator tasks to test (comma-separated)')
@click.option('--workflows', type=str, default='', help='Secator workflows to test (comma-separated)')
@click.option('--scans', type=str, default='', help='Secator scans to test (comma-separated)')
@click.option('--test', '-t', type=str, help='Secator test to run')
def integration(tasks, workflows, scans, test):
	"""Run integration tests."""
	os.environ['TEST_TASKS'] = tasks or ''
	os.environ['TEST_WORKFLOWS'] = workflows or ''
	os.environ['TEST_SCANS'] = scans or ''
	os.environ['SECATOR_DIRS_DATA'] = '/tmp/.secator'
	os.environ['SECATOR_RUNNERS_SKIP_CVE_SEARCH'] = '1'

	import shutil
	shutil.rmtree('/tmp/.secator', ignore_errors=True)

	cmd = f'{sys.executable} -m coverage run --omit="*test*" --data-file=.coverage.integration -m pytest -s -v tests/integration'  # noqa: E501
	if test:
		test_str = ' or '.join(test.split(','))
		cmd += f' -k "{test_str}"'
	run_test(cmd, 'integration')


@test.command()
@click.option('--tasks', type=str, default='', help='Secator tasks to test (comma-separated)')
@click.option('--workflows', type=str, default='', help='Secator workflows to test (comma-separated)')
@click.option('--scans', type=str, default='', help='Secator scans to test (comma-separated)')
@click.option('--test', '-t', type=str, help='Secator test to run')
def performance(tasks, workflows, scans, test):
	"""Run integration tests."""
	os.environ['TEST_TASKS'] = tasks or ''
	os.environ['TEST_WORKFLOWS'] = workflows or ''
	os.environ['TEST_SCANS'] = scans or ''
	os.environ['SECATOR_DIRS_DATA'] = '/tmp/.secator'
	os.environ['SECATOR_RUNNERS_SKIP_CVE_SEARCH'] = '1'

	# import shutil
	# shutil.rmtree('/tmp/.secator', ignore_errors=True)

	cmd = f'{sys.executable} -m coverage run --omit="*test*" --data-file=.coverage.performance -m pytest -s -v tests/performance'  # noqa: E501
	if test:
		test_str = ' or '.join(test.split(','))
		cmd += f' -k "{test_str}"'
	run_test(cmd, 'performance')


@test.command()
@click.option('--unit-only', '-u', is_flag=True, default=False, help='Only generate coverage for unit tests')
@click.option('--integration-only', '-i', is_flag=True, default=False, help='Only generate coverage for integration tests')  # noqa: E501
def coverage(unit_only, integration_only):
	"""Run coverage combine + coverage report."""
	cmd = f'{sys.executable} -m coverage report -m --omit=*/site-packages/*,*/tests/*,*/templates/*'
	if unit_only:
		cmd += ' --data-file=.coverage.unit'
	elif integration_only:
		cmd += ' --data-file=.coverage.integration'
	else:
		Command.execute(f'{sys.executable} -m coverage combine --keep', name='coverage combine', cwd=ROOT_FOLDER)
	run_test(cmd, 'coverage')



import time
from dataclasses import dataclass, field
from secator.output_types import OutputType
from secator.utils import rich_to_ansi, rich_escape as _s
from secator.definitions import MATCHED_AT, NAME, ID, EXTRA_DATA, REFERENCE


@dataclass
class Exploit(OutputType):
	name: str
	provider: str
	id: str
	matched_at: str = ''
	ip: str = ''
	confidence: str = 'low'
	cvss_score: float = 0
	reference: str = ''
	cves: list = field(default_factory=list, compare=False)
	tags: list = field(default_factory=list, compare=False)
	extra_data: dict = field(default_factory=dict, compare=False)
	_source: str = field(default='', repr=True)
	_type: str = field(default='exploit', repr=True)
	_timestamp: int = field(default_factory=lambda: time.time(), compare=False)
	_uuid: str = field(default='', repr=True, compare=False)
	_context: dict = field(default_factory=dict, repr=True, compare=False)
	_tagged: bool = field(default=False, repr=True, compare=False)
	_duplicate: bool = field(default=False, repr=True, compare=False)
	_related: list = field(default_factory=list, compare=False)

	_table_fields = [
		MATCHED_AT,
		NAME,
		ID,
		EXTRA_DATA,
		REFERENCE
	]
	_sort_by = ('matched_at', 'name')

	def __str__(self):
		return self.name

	def __repr__(self):
		s = rf'[bold red]⍼[/]  \[[bold red]{self.name}'
		if self.reference:
			s += f' [link={_s(self.reference)}]🡕[/link]'
		s += '[/]]'
		if self.matched_at:
			s += f' {_s(self.matched_at)}'
		if self.tags:
			tags_str = ', '.join(self.tags)
			s += rf' \[[cyan]{tags_str}[/]]'
		if self.extra_data:
			data = ', '.join([f'{k}:{v}' for k, v in self.extra_data.items()])
			s += rf' \[[yellow]{_s(str(data))}[/]]'
		if self.confidence == 'low':
			s = f'[dim]{s}[/]'
		return rich_to_ansi(s)



from secator.exporters._base import Exporter
from secator.utils import pluralize
from secator.rich import build_table, console
from rich.markdown import Markdown
from secator.output_types import OutputType


class TableExporter(Exporter):
	def send(self):
		results = self.report.data['results']
		if not results:
			return
		title = self.report.title
		_print = console.print
		_print()
		if title:
			title = ' '.join(title.capitalize().split('_')) + ' results'
			h1 = Markdown(f'# {title}')
			_print(h1, style='bold magenta', width=50)
			_print()
		for output_type, items in results.items():
			if output_type == 'progress':
				continue
			if items:
				is_output_type = isinstance(items[0], OutputType)
				output_fields = items[0]._table_fields if is_output_type else None
				sort_by = items[0]._sort_by if is_output_type else []
				_table = build_table(
					items,
					output_fields=output_fields,
					sort_by=sort_by)
				title = pluralize(items[0]._type).upper() if is_output_type else 'Results'
				_print(f':wrench: {title}', style='bold gold3', justify='left')
				_print(_table)
				_print()



import time
from dataclasses import dataclass, field

from secator.definitions import CPES, EXTRA_DATA, HOST, IP, PORT
from secator.output_types import OutputType
from secator.utils import rich_to_ansi


@dataclass
class Port(OutputType):
	port: int
	ip: str
	state: str = 'UNKNOWN'
	service_name: str = field(default='', compare=False)
	cpes: list = field(default_factory=list, compare=False)
	host: str = field(default='', repr=True, compare=False)
	protocol: str = field(default='tcp', repr=True, compare=False)
	extra_data: dict = field(default_factory=dict, compare=False)
	confidence: str = field(default='low', repr=False, compare=False)
	_timestamp: int = field(default_factory=lambda: time.time(), compare=False)
	_source: str = field(default='', repr=True, compare=False)
	_type: str = field(default='port', repr=True)
	_uuid: str = field(default='', repr=True, compare=False)
	_context: dict = field(default_factory=dict, repr=True, compare=False)
	_tagged: bool = field(default=False, repr=True, compare=False)
	_duplicate: bool = field(default=False, repr=True, compare=False)
	_related: list = field(default_factory=list, compare=False)

	_table_fields = [IP, PORT, HOST, CPES, EXTRA_DATA]
	_sort_by = (PORT, IP)

	def __gt__(self, other):
		# favor nmap over other port detection tools
		if self._source == 'nmap' and other._source != 'nmap':
			return True
		return super().__gt__(other)

	def __str__(self) -> str:
		return f'{self.host}:{self.port}'

	def __repr__(self) -> str:
		s = f'🔓 {self.ip}:[bold red]{self.port:<4}[/] [bold yellow]{self.state.upper()}[/]'
		if self.protocol != 'TCP':
			s += rf' \[[yellow3]{self.protocol}[/]]'
		if self.service_name:
			conf = ''
			if self.confidence == 'low':
				conf = '?'
			s += rf' \[[bold purple]{self.service_name}{conf}[/]]'
		if self.host:
			s += rf' \[[cyan]{self.host}[/]]'
		return rich_to_ansi(s)



from secator.exporters._base import Exporter
from secator.output_types import Info
from secator.rich import console
from secator.serializers.dataclass import dumps_dataclass


class JsonExporter(Exporter):
	def send(self):
		json_path = f'{self.report.output_folder}/report.json'

		# Save JSON report to file
		with open(json_path, 'w') as f:
			f.write(dumps_dataclass(self.report.data, indent=2))

		info = Info(f'Saved JSON report to {json_path}')
		console.print(info)



import unittest
from unittest.mock import patch, MagicMock
from io import StringIO
import sys
from secator.runners import Command
from secator.runners._base import HOOKS
from secator.utils_test import mock_command
from secator.output_types import OutputType, Url, Vulnerability, Tag, Target
from secator.serializers.regex import RegexSerializer
from secator.serializers.json import JSONSerializer


class MyCommand(Command):
	cmd = 'dummy'
	input_flag = '-u'
	file_flag = None


TARGETS = ['host1']
FIXTURE = ['test_fixture']


class TestCommandRunner(unittest.TestCase):

	def setUp(self):
		self.original_output_types = MyCommand.output_types
		self.original_item_loaders = MyCommand.item_loaders
		self.all_hooks = list(set(Command.hooks + HOOKS))

	def tearDown(self):
		MyCommand.output_types = self.original_output_types
		MyCommand.item_loaders = self.original_item_loaders
		self.cleanup_hooks()

	def mock_hooks(self, output_types=[], item_loaders=[]):
		"""Set up command for testing with customizable parameters."""
		MyCommand.output_types = output_types
		MyCommand.item_loaders = item_loaders

		mock_hooks = {}
		for hook in self.all_hooks:
			if hook in ['on_item_pre_convert', 'on_item', 'on_line', 'on_duplicate']:
				def func(self, item):
					return item
			else:
				def func(self):
					return None
			func = staticmethod(func)
			mock = self.mock_hook(hook, func)
			mock_hooks[hook] = mock

		# Dynamic hook generation for all serializers
		for loader in MyCommand.item_loaders:
			serializer_name = loader.__class__.__name__.lower().replace('serializer', '')
			hook = f'on_{serializer_name}_loaded'
			self.all_hooks.append(hook)
			def mock_generator(self, item):
				yield item
			mock = self.mock_hook(hook, mock_generator)
			mock_hooks[hook] = mock

		return mock_hooks

	def mock_hook(self, hook_name, hook_func):
		mock = MagicMock()
		mock.call_count = 0
		mock.__name__ = hook_name
		mock.__qualname__ = hook_name
		mock.side_effect = hook_func
		setattr(MyCommand, hook_name, mock)
		return mock

	def cleanup_hooks(self):
		"""Clean up hooks after testing."""
		for hook in self.all_hooks:
			if hasattr(MyCommand, hook):
				delattr(MyCommand, hook)

	def test_init(self):
		"""Ensure that init the command correctly inits the Runner and calls the init hooks before_init and on_init"""
		def before_init(self):
			pass

		def on_init(self):
			pass

		MyCommand.before_init = staticmethod(before_init)
		MyCommand.on_init = staticmethod(on_init)

		with patch.object(Command, 'run_hooks') as mock_run_hooks:
			MyCommand(TARGETS)
			mock_run_hooks.assert_any_call('before_init')
			mock_run_hooks.assert_any_call('on_init')

		# Clean up after test
		delattr(MyCommand, 'before_init')
		delattr(MyCommand, 'on_init')

	def test_opts(self):
		"""Ensure that opts are correctly defined and mapped to the right output options."""
		MyCommand.opts = {
			'test_flag': {'is_flag': True, 'default': False, 'help': 'Test flag option'},
			'test_str': {'is_flag': False, 'default': None, 'help': 'Test string option'}
		}
		MyCommand.opt_key_map = {'test_flag': 'tf', 'test_str': 'ts'}
		self.all_hooks.extend(['opts', 'opt_key_map'])

		test_opts = {'test_flag': True, 'test_str': 'test_value'}
		cmd = MyCommand(TARGETS, **test_opts)
		self.assertEqual(cmd.run_opts['test_flag'], True)
		self.assertEqual(cmd.run_opts['test_str'], 'test_value')

		# Check if the command is correctly constructed
		expected_cmd = 'dummy -u host1 -tf -ts test_value'
		self.assertEqual(cmd.cmd, expected_cmd)


	def test_hooks(self):
		"""Ensure that all hooks in the Command + Runner lifecycle are correctly called."""
		mock_hooks = self.mock_hooks(
			output_types=[Url],
			item_loaders=[JSONSerializer()],
		)
		fixture=[
			'{"url": "http://example.com"}',
			'{"url": "http://example.org"}',
			'{"url": "http://example.com"}'  # Duplicate
		]

		# Run the command using mock_command
		with mock_command(MyCommand, TARGETS, {}, fixture, 'run'):
			for hook, mock in mock_hooks.items():
				self.assertTrue(mock.called, f"Hook '{hook}' was not called")
			self.assertEqual(mock_hooks['on_json_loaded'].call_count, 3)
			self.assertGreaterEqual(mock_hooks['on_duplicate'].call_count, 1)
			for item_hook in ['on_item_pre_convert', 'on_item', 'on_line', 'on_duplicate', 'on_json_loaded']:
				for call in mock_hooks[item_hook].call_args_list:
					self.assertEqual(len(call[0]), 2)  # self and item arguments
					self.assertIsInstance(call[0][1], (str, dict, OutputType))  # result

	def test_hooks_failing(self):
		"""Ensure that failing hooks are correctly handled for all hooks in the command lifecycle."""
		fixture = [
			'{"url": "http://example.com"}',
			'{"url": "http://example.org"}',
			'{"url": "http://example.com"}'  # Duplicate
		]
		MyCommand.output_types = [Url]
		MyCommand.item_loaders = [JSONSerializer()]
		all_hooks = self.all_hooks
		for failing_hook in all_hooks:
			with self.subTest(failing_hook=failing_hook):
				def failing_hook_func(self, *args, **kwargs):
					raise Exception(f"Hook {failing_hook} failed")
				self.mock_hook(failing_hook, failing_hook_func)
				captured_output = StringIO()
				sys.stderr = captured_output
				with mock_command(MyCommand, TARGETS, {}, fixture) as command:
					command.run()
					errors = [e.message for e in command.errors]
					if errors:  # error happened during the actual execution, it will be yielded in results
						self.assertIn(f'Hook "unittest.mock.{failing_hook}" execution failed.', errors)
						self.assertEqual(command.status, 'FAILURE')
				delattr(MyCommand, failing_hook)

	def test_input_loaders(self):
		"""Verify what happens when one of the input loaders fail."""
		def failing_loader(self, line):
			raise Exception("Loader failed")

		MyCommand.item_loaders = [failing_loader]
		with mock_command(MyCommand, TARGETS, {}, FIXTURE) as command:
			results = command.run()
			errors = [e.message for e in results if e._type == 'error']
			self.assertIn('Exception: Loader failed', errors)
			self.assertEqual(len(command.results), 2)
			self.assertEqual(command.status, 'FAILURE')

	def test_inputs_validator_failed_no_targets(self):
		cmd = MyCommand([])
		cmd.run()
		errors = cmd.errors
		messages = [e.message for e in errors]
		self.assertIn("Validator failed: Input is empty.", messages)
		self.assertEqual(len(cmd.results), 1)
		self.assertFalse(cmd.inputs_valid)
		self.assertEqual(cmd.status, 'FAILURE')

	def test_inputs_validator_failed_multiple_targets(self):
		targets = ['host1', 'host2']
		cmd = MyCommand(targets)
		cmd.run()
		errors = cmd.errors
		messages = [e.message for e in errors]
		self.assertIn("Validator failed: Command does not suport multiple inputs in non-worker mode. Consider using .delay() instead.", messages)
		self.assertEqual(len(cmd.results), 1)
		self.assertFalse(cmd.inputs_valid)
		self.assertEqual(cmd.status, 'FAILURE')

	# def test_inputs_validator_failed_wrong_input_type(self):
	# 	MyCommand.input_types = [Url]
	# 	targets = ['host1', 'host2'] 
	# 	cmd = MyCommand(targets)
	# 	errors = cmd.errors
	# 	messages = [e.message for e in errors]
	# 	self.assertIn("Validator failed: Multiple input passed in non-worker mode.", messages)
	# 	self.assertEqual(len(cmd.results), 0)
	# 	self.assertFalse(cmd.inputs_valid)
	# 	delattr(MyCommand, 'input_types', None)

	def test_stdout_output(self):
		"""Test MyCommand with stdout output."""
		fixture = [
			"http://example.com | URL | example.com | 200 | 1000 | Example Title",
			"CVE-2021-1234 | VULNERABILITY | CVE-2021-1234 | http://example.com",
			"AWS_KEY | TAG | AKIA1234567890ABCDEF | http://example.com"
		]
		MyCommand.output_types = [Url, Tag, Vulnerability]
		MyCommand.item_loaders = [
			RegexSerializer(
				r'^(?P<url>.*?) \| (?P<type>.*?) \| (?P<host>.*?) \| (?P<status_code>\d+) \| (?P<content_length>\d+)( \| (?P<title>.*))?$',
				fields=['url', 'type', 'host', 'status_code', 'content_length', 'title']
			),
			RegexSerializer(
				r'^(?P<name>.*?) \| VULNERABILITY \| (?P<id>.*?) \| (?P<matched_at>.*?)$',
				fields=['name', 'id', 'matched_at']
			),
			RegexSerializer(
				r'^(?P<name>.*?) \| TAG \| (?P<match>.*?) \| (?P<matched_at>.*?)$',
				fields=['name', 'match', 'matched_at']
			)
		]
		def on_regex_loaded(self, item):
			if 'status_code' in item:
				item['status_code'] = int(item['status_code'])
			yield item

		MyCommand.on_regex_loaded = staticmethod(on_regex_loaded)
		self.all_hooks.extend(['output_types', 'item_loaders', 'on_regex_loaded'])

		with mock_command(MyCommand, TARGETS, {}, fixture, 'run') as results:
			results = list(results)
			self.assertEqual(len(results), 4)
			self.assertIsInstance(results[0], Target)
			self.assertIsInstance(results[1], Url)
			self.assertIsInstance(results[2], Vulnerability)
			self.assertIsInstance(results[3], Tag)

			# Check specific attributes
			self.assertEqual(results[0].name, "host1")
			self.assertEqual(results[1].url, "http://example.com")
			self.assertEqual(results[2].id, "CVE-2021-1234")
			self.assertEqual(results[3].name, "AWS_KEY")

	def test_jsonlines_output(self):
		"""Test MyCommand with JSON-lines output."""
		fixture = [
			'{"url": "http://host1:5000/api/", "status": 200}',
			'{"vulnerability": "myvuln", "severity": "HIGH", "matched": "http://host1:5000/api/"}',
			'{"tag": "mytag", "matched_at": "http://host1:5000/api/", "tag_type": "AWS_API_KEY", "value": "ACDIFOJ-ASDF"}'
		]
		MyCommand.output_types = [Url, Vulnerability, Tag]
		MyCommand.item_loaders = [JSONSerializer()]
		MyCommand.output_map = {
			Url: {
				'url': 'url',
				'status_code': 'status'
			},
			Vulnerability: {
				'name': 'vulnerability',
				'severity': 'severity',
				'matched_at': 'matched'
			},
			Tag: {
				'name': 'tag',
				'match': 'matched_at',
				'extra_data': lambda x: {k: v for k, v in x.items() if k in ['tag_type', 'value']}
			}
		}
		self.all_hooks.extend(['output_types', 'item_loaders', 'output_map'])

		with mock_command(MyCommand, TARGETS, {}, fixture, 'run') as results:
			results = list(results)
			self.assertEqual(len(results), 4)
			self.assertIsInstance(results[0], Target)
			self.assertIsInstance(results[1], Url)
			self.assertIsInstance(results[2], Vulnerability)
			self.assertIsInstance(results[3], Tag)

			# Check specific attributes
			self.assertEqual(results[0].name, "host1")
			self.assertEqual(results[1].url, "http://host1:5000/api/")
			self.assertEqual(results[2].name, "myvuln")
			self.assertEqual(results[3].name, "mytag")
			self.assertEqual(results[3].extra_data['tag_type'], "AWS_API_KEY")

	def test_json_file_output(self):
		"""Test MyCommand with JSON file output."""
		json_output = [
			{"url": "http://example.com", "status_code": 200},
			{"name": "SQL Injection", "severity": "high", "matched_at": "http://example.com"},
			{"name": "sensitive_data", "match": "http://example.com", "extra_data": {"tag_type": "PII", "value": "SSN"}}
		]

		import json
		with open('output.json', 'w') as f:
			json.dump(json_output, f)

		def on_cmd_done(self):
			with open('output.json', 'r') as f:
				data = json.load(f)

			for item in data:
				if 'url' in item:
					yield Url(**item)
				elif 'match' in item:
					yield Tag(**item)
				else:
					yield Vulnerability(**item)

		MyCommand.on_cmd_done = staticmethod(on_cmd_done)
		self.all_hooks.extend(['on_cmd_done'])

		with mock_command(MyCommand, TARGETS, {}, [], 'run') as results:
			self.assertEqual(len(results), 4)
			self.assertIsInstance(results[0], Target)
			self.assertIsInstance(results[1], Url)
			self.assertIsInstance(results[2], Vulnerability)
			self.assertIsInstance(results[3], Tag)
	
			# Check specific attributes
			self.assertEqual(results[0].name, "host1")
			self.assertEqual(results[1].url, "http://example.com")
			self.assertEqual(results[2].name, "SQL Injection")
			self.assertEqual(results[3].name, "sensitive_data")

		# Clean up the temporary file
		import os
		os.remove('output.json')

	def test_convert_item_schema(self):
		MyCommand.output_types = [Url, Tag, Vulnerability]
		url = TARGETS[0]
		items = [
			{
				'__test__': 'Item with _type hint should load properly and bypass other output types',
				'url': url,
				'_type': 'url',
				'__expected__': {
					'_type': 'url',
				},
			},
			{
				'__test__': 'Item with all the required fields should load properly',
				'url': url,
				'__expected__': {
					'_type': 'url',
				},
			},
			{
				'__test__': 'Items with all the required fields AND fields outside schema should load properly',
				'url': url,
				'extra_field_not_in_schema': {'new': 'added', 'field': 'old'},
				'__expected__': {
					'_type': 'url',
				},
			},
			{
				'__test__': 'Items with missing required fields should fail to load',
				'target': url,
				'__expected__': {
					'_type': 'warning',
				},
			},
			{
				'__test__': 'Items with missing required fields AND _type hint should fail to load',
				'_type': 'vulnerability',
				'__expected__': {
					'_type': 'warning',
				},
			},
			{
				'__test__': 'Item with all the required fields should load properly',
				'name': 'SQL Injection',
				'severity': 'high',
				'matched_at': 'http://example.com',
				'__expected__': {
					'_type': 'vulnerability'
				},
			},
			{
				'__test__': ' Trying to load a Vulnerability as a Tag should output a warning',
				"name": "SQL Injection",
				"severity": "high",
				"matched_at": "http://example.com",
				"_type": "tag",
				'__expected__': {
					'_type': 'warning'
				},
			},
			{
				'__test__': 'Item of type vulnerability with no _type hint should be incorrectly loaded as Tag',
				'name': 'sensitive_data',
				'match': 'http://example.com',
				'extra_data': {
					'tag_type': 'PII',
					'value': 'SSN'
				},
				'__expected__': {
					'_type': 'tag'
				},
			},
			{
				'__test__': 'Item of type vulnerability with _type hint will be loaded properly',
				'name': 'sensitive_data',
				'match': 'http://example.com',
				'extra_data': {
					'tag_type': 'PII',
					'value': 'SSN'
				},
				'_type': 'vulnerability',
				'__expected__': {
					'_type': 'vulnerability',
				}
			}
		]
		with mock_command(MyCommand, TARGETS, {}, []) as cmd:
			for item in items:
				msg = item.pop('__test__')
				expected_fields = item.pop('__expected__')
				converted = cmd._convert_item_schema(item)
				with self.subTest(msg=msg, item=item, converted=converted.toDict()):
					for k, v in expected_fields.items():
						self.assertEqual(getattr(converted, k), v)
		delattr(MyCommand, 'output_types')



__all__ = [
	'ConsoleExporter',
	'CsvExporter',
	'GdriveExporter',
	'JsonExporter',
	'TableExporter',
	'TxtExporter'
]
from secator.exporters.console import ConsoleExporter
from secator.exporters.csv import CsvExporter
from secator.exporters.gdrive import GdriveExporter
from secator.exporters.json import JsonExporter
from secator.exporters.table import TableExporter
from secator.exporters.txt import TxtExporter



import json
import unittest

from secator.utils_test import load_fixture, FIXTURES_DIR
from secator.tasks._categories import Vuln
from secator.config import CONFIG


class TestSerializers(unittest.TestCase):

	def test_lookup_cve_circle(self):
		fixture = json.dumps(load_fixture('cve_circle_output', FIXTURES_DIR), sort_keys=True)
		actual = json.dumps(Vuln.lookup_cve_from_cve_circle('CVE-2023-5568'), sort_keys=True)
		self.assertEqual(actual, fixture)

	def test_lookup_cve(self):
		fixture = load_fixture('cve_circle_output', FIXTURES_DIR)
		cve_path = f'{CONFIG.dirs.data}/cves/CVE-2023-5568.json'
		with open(cve_path, 'w') as f:
			f.write(json.dumps(fixture, indent=2))
		vuln = Vuln.lookup_cve('CVE-2023-5568', 'cpe:/o:redhat:enterprise_linux:9')
		vuln2 = Vuln.lookup_cve('CVE-2023-5568', 'cpe:2.3:o:redhat:enterprise_linux:9:*:*:*:*:*:*:*')
		self.assertIn('cpe-match', vuln['tags'])
		self.assertIn('cpe-match', vuln2['tags'])



from pathlib import Path
from time import time

from google.cloud import storage

from secator.config import CONFIG
from secator.runners import Task
from secator.thread import Thread
from secator.utils import debug


GCS_BUCKET_NAME = CONFIG.addons.gcs.bucket_name
ITEMS_TO_SEND = {
	'url': ['screenshot_path']
}


def process_item(self, item):
	if item._type not in ITEMS_TO_SEND.keys():
		return item
	if not GCS_BUCKET_NAME:
		debug('skipped since addons.gcs.bucket_name is empty.', sub='hooks.gcs')
		return item
	to_send = ITEMS_TO_SEND[item._type]
	for k, v in item.toDict().items():
		if k in to_send and v:
			path = Path(v)
			if not path.exists():
				continue
			ext = path.suffix
			blob_name = f'{item._uuid}_{k}{ext}'
			t = Thread(target=upload_blob, args=(GCS_BUCKET_NAME, v, blob_name))
			t.start()
			self.threads.append(t)
			setattr(item, k, f'gs://{GCS_BUCKET_NAME}/{blob_name}')
	return item


def upload_blob(bucket_name, source_file_name, destination_blob_name):
	"""Uploads a file to the bucket."""
	start_time = time()
	storage_client = storage.Client()
	bucket = storage_client.bucket(bucket_name)
	blob = bucket.blob(destination_blob_name)
	blob.upload_from_filename(source_file_name)
	end_time = time()
	elapsed = end_time - start_time
	debug(f'in {elapsed:.4f}s', obj={'blob': 'CREATED', 'blob_name': destination_blob_name, 'bucket': bucket_name}, obj_after=False, sub='hooks.gcs', verbose=True)  # noqa: E501


HOOKS = {
	Task: {'on_item': [process_item]}
}



import os
import shutil
import unittest
import yaml

from pathlib import Path
from unittest import mock

devnull = open(os.devnull, 'w')

@mock.patch('sys.stderr', devnull)
class TestConfig(unittest.TestCase):

	def setUp(self):
		self.env_dirs_data = os.environ['SECATOR_DIRS_DATA']
		del os.environ['SECATOR_DIRS_DATA']
		self.home = Path.home()
		self.valid_config = {
			'addons': {'gdrive': {'enabled': True}}
		}
		self.invalid_config = {
			'addons': {'gdrive': {'enabled': 'non-boolean'}}
		}
		self.config_home_dir = {
			'dirs': {'data': '~/test'}
		}
		self.config_home_dir_reduce = {
			'dirs': {'data': f'{self.home}/test'}
		}
		self.config_test = Path('test.yml')
		self.config_test.touch()

	def tearDown(self):
		self.config_test.unlink()
		os.environ['SECATOR_DIRS_DATA'] = self.env_dirs_data

	def test_parse_empty_config(self):
		from secator.config import Config
		config = Config.parse()
		self.assertIsNotNone(config)
		self.assertIsInstance(config, Config)
		self.assertIsInstance(config.addons.gdrive.enabled, bool)

	def test_parse_valid_config(self):
		from secator.config import Config
		config = Config.parse(self.valid_config)
		self.assertIsNotNone(config)
		self.assertIsInstance(config, Config)
		self.assertEqual(config.addons.gdrive.enabled, True)

	def test_parse_invalid_config(self):
		from secator.config import Config
		config = Config.parse(self.invalid_config)
		self.assertIsNone(config)

	def test_parse_home_dir_expand(self):
		from secator.config import Config
		user_data_dir = self.config_home_dir['dirs']['data']
		config = Config.parse(self.config_home_dir)
		self.assertEqual(config.dirs.data, Path(user_data_dir).expanduser())
		config.save(self.config_test)
		config = Config.parse(path=self.config_test)
		self.assertEqual(config.dirs.data, Path(user_data_dir).expanduser())

	def test_set_config_key(self):
		from secator.config import Config
		config = Config.parse(path=self.config_test)
		config.set('addons.gdrive.enabled', True)
		config.save()
		yaml_data = Config.read_yaml(self.config_test)
		self.assertEqual(yaml_data['addons']['gdrive']['enabled'], True)
		self.assertEqual(config.get('addons.gdrive.enabled'), True)
		config = Config.parse(path=self.config_test)
		self.assertEqual(config.addons.gdrive.enabled, True)
		self.assertEqual(config._partial.addons.gdrive.enabled, True)

	def test_parse_home_dir_reduce(self):
		from secator.config import Config
		with self.config_test.open('w') as f:
			f.write(yaml.dump(self.config_home_dir_reduce))
		config = Config.parse(path=self.config_test)
		self.assertIsInstance(config.dirs.data, Path)
		self.assertIsInstance(config._partial.dirs.data, str)
		config.save(self.config_test)
		data = Config.read_yaml(self.config_test)
		self.assertNotIn(str(self.home), data['dirs']['data'])


@mock.patch('sys.stderr', devnull)
class TestConfigEnv(unittest.TestCase):

	def setUp(self):
		from secator.utils_test import clear_modules
		clear_modules()
		shutil.rmtree('/tmp/.secator', ignore_errors=True)
		Path('/tmp/.secator').mkdir(parents=False)

	def tearDown(self):
		shutil.rmtree('/tmp/.secator', ignore_errors=True)

	@mock.patch.dict(os.environ, {'SECATOR_DIRS_DATA': '/tmp/.secator/new'})
	def test_dirs_data(self):
		from secator.config import CONFIG
		self.assertEqual(CONFIG.dirs.data, Path('/tmp/.secator/new'))

		# Check other dirs configured OK
		ignore_dirs = ['bin', 'share', 'data']
		for k, dir in CONFIG.dirs.items():
			if k in ignore_dirs:
				continue
			rel_target = '/'.join(k.split('_'))
			expected_path = CONFIG.dirs.data / rel_target
			self.assertEqual(dir, expected_path)

		# Check all dirs exist
		for dir in CONFIG.dirs.values():
			self.assertTrue(dir.exists())

	@mock.patch.dict(os.environ, {'SECATOR_DIRS_DATA': '/tmp/.secator/new', 'SECATOR_DIRS_PAYLOADS': '/tmp/.secator/payloads2'})
	def test_dirs_data_and_others(self):
		from secator.config import CONFIG
		self.assertEqual(CONFIG.dirs.data, Path('/tmp/.secator/new'))
		self.assertEqual(CONFIG.dirs.payloads, Path('/tmp/.secator/payloads2'))

		# Check other dirs configured OK
		ignore_dirs = ['bin', 'share', 'data', 'payloads']
		for k, dir in CONFIG.dirs.items():
			if k in ignore_dirs:
				continue
			rel_target = '/'.join(k.split('_'))
			expected_path = CONFIG.dirs.data / rel_target
			self.assertEqual(dir, expected_path)

		# Check all dirs exist
		for dir in CONFIG.dirs.values():
			self.assertTrue(dir.exists())

	@mock.patch.dict(os.environ, {'SECATOR_DIRS_PAYLOADS': '/tmp/.secator/payloads2', 'SECATOR_DIRS_TEMPLATES': '/tmp/.secator/templates2'})
	def test_dirs_others_only(self):
		from secator.config import CONFIG
		self.assertEqual(CONFIG.dirs.data, Path('/tmp/.secator'))
		self.assertEqual(CONFIG.dirs.payloads, Path('/tmp/.secator/payloads2'))
		self.assertEqual(CONFIG.dirs.templates, Path('/tmp/.secator/templates2'))

		# Check other dirs configured OK
		ignore_dirs = ['bin', 'share', 'data', 'payloads', 'templates']
		for k, dir in CONFIG.dirs.items():
			if k in ignore_dirs:
				continue
			rel_target = '/'.join(k.split('_'))
			expected_path = CONFIG.dirs.data / rel_target
			self.assertEqual(dir, expected_path)

		# Check all dirs exist
		for dir in CONFIG.dirs.values():
			self.assertTrue(dir.exists())
